package edu.lehigh.swat.vpi;

/**
 * <p>Title:  ???</p>
 * <p>Description: Virtual Perspective Interface on Jena. This is a query system
 *                 for querying a group of backward compatible ontologies. The
 *                 query can be passed on any perspective (one of the ontologies).
 *                 The system will automatically run a search for relevant results
 *                 based on the query on all backward compatible previous versions
 *                 and give out results with corresponding name spaces
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: LU </p>
 * @author Z. Pan and Upmanyu Misra
 * @version 1.0
 */

import com.hp.hpl.jena.rdf.model.*;
import com.hp.hpl.jena.graph.*;
import com.hp.hpl.jena.rdql.*;
import com.hp.hpl.jena.ontology.*;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;
import com.hp.hpl.jena.vocabulary.RDFS;
import com.hp.hpl.jena.vocabulary.RDF;
import com.hp.hpl.jena.vocabulary.OWL;
import com.hp.hpl.jena.rdf.model.impl.SelectorImpl;
import java.util.*;
import java.io.*;
import java.lang.*;

public class VPI {

  /**
   * Constructor
   */
  public VPI() {

    documentManager = OntDocumentManager.getInstance();
    //documentManager.setProcessImports(false); // it doesn't handle imports right now
    loadedDocuments = new Hashtable();
    globalModel = ModelFactory.createOntologyModel(OntModelSpec.
        OWL_MEM_RDFS_INF
        , null);
    hasVP = new Vector();

  }

  /**
   * Load a document into the VPI.
   * @param uri
   */
  public void loadDocument(String uri) {
    try
    {
      //model for the document being loaded
      StringTokenizer str = new StringTokenizer(uri, "/");
      String pre = "";
      while (str.hasMoreTokens())
      {
        pre = str.nextToken();
      }

      OntModel model = ModelFactory.createOntologyModel(OntModelSpec.
          OWL_MEM
          , null);
      model.read(uri);
      loadedDocuments.put(uri, model);

      globalModel.add(model);

      Ontology ont = model.getOntology(uri);

      //selecting backward-compatible ontologies of the current document
      Hashtable bcwOntos = new Hashtable();
      if (ont != null)
      {
        ExtendedIterator ei = ont.listBackwardCompatibleWith();
        while (ei.hasNext())
        {
          OntResource bcw = (OntResource) ei.next();
          String bcw_uri = bcw.getURI();
          if (!loadedDocuments.contains(bcw_uri))
          {
            this.loadDocument(bcw_uri);
          }
          bcwOntos.put(bcw_uri, loadedDocuments.get(bcw_uri));
        }
      }
      for (Enumeration en = bcwOntos.elements(); en.hasMoreElements(); )
      {
        OntModel bcwo = (OntModel) en.nextElement();
        globalModel.add(bcwo);
      }

      //model for the virtual perspective ontology
      OntModel vpModel = ModelFactory.createOntologyModel(OntModelSpec.
          OWL_MEM_RDFS_INF
          , null);

      //model for the Type Graph ontology
      OntModel typModel = ModelFactory.createOntologyModel(OntModelSpec.
          OWL_MEM_RDFS_INF
          , null);

      //copy all the classes in the current model to a vector
      ExtendedIterator cei = model.listNamedClasses();
      Vector classes = new Vector();
      while (cei.hasNext()) {
        OntClass cls = (OntClass) cei.next();
        classes.add(cls);
      }

      //copy all the properties in the current model to a vector
      ExtendedIterator pei = model.listObjectProperties();
      Vector properties = new Vector();
      while (pei.hasNext())
      {
        OntProperty prop = (OntProperty) pei.next();
        properties.add(prop);
      }
      pei = model.listDatatypeProperties();
      while (pei.hasNext())
      {
        OntProperty prop = (OntProperty) pei.next();
        properties.add(prop);
      }

      //iterate the classes in the current model, create classes in vpModel
      //if a BCW model also has this class, set up subClassOf relations
      Iterator it = classes.iterator();
      while (it.hasNext())
      {
        OntClass cls = (OntClass) it.next();
        if (!cls.isAnon())
        {
          String localName = cls.getLocalName();
          String ns = cls.getNameSpace();
          String vns = "VPI://".concat(ns);
          OntClass vcls = vpModel.createClass(vns + localName);
          Resource ocls = vpModel.createResource(ns + localName);
          Statement s = vpModel.createStatement(ocls, RDFS.subClassOf, vcls);
          vpModel.add(s);

          String tns = "TYP://".concat(ns);
          OntClass tcls = typModel.createClass(tns + localName);
          Statement ts = typModel.createStatement(tcls, RDFS.subClassOf, vcls);
          typModel.add(ts);

          for (Enumeration en = bcwOntos.keys(); en.hasMoreElements(); )
          {
            String bns = (String) en.nextElement();
            OntClass bcls = globalModel.getOntClass(bns + "#" + localName);
            if (bcls != null)
            {
              Resource tbc = typModel.createResource("TYP://" + bns + "#" +
                  localName);
              Statement ts2 = typModel.createStatement(tbc, RDFS.subClassOf,
                  tcls);
              typModel.add(ts2);
            }
          }
        }
      }

      /*iterate the properties in the current model, create properties in vpModel
        and the type model
      */
      //if a BCW model also has this property, set up subPropertyOf relations
      Iterator pit = properties.iterator();
      while (pit.hasNext())
      {
        OntProperty prop = (OntProperty) pit.next();
        if (!prop.isAnon())
        {
          String localName = prop.getLocalName();
          String ns = prop.getNameSpace();
          String vns = "VPI://".concat(ns);
          String tns = "TYP://".concat(ns);
          OntProperty vprop = vpModel.createOntProperty(vns + localName);
          OntProperty tprop = typModel.createOntProperty(tns + localName);
          Resource oprop = vpModel.createResource(ns + localName);

          Statement s = vpModel.createStatement(oprop, RDFS.subPropertyOf,
                                                vprop);
          vpModel.add(s);
          Statement s0 = vpModel.createStatement(tprop, RDFS.subPropertyOf,
                                                 vprop);
          typModel.add(s0);
          for (Enumeration en = bcwOntos.keys(); en.hasMoreElements(); )
          {
            String bns = (String) en.nextElement();
            OntProperty bprop = globalModel.getOntProperty(bns + "#" +
                localName);
            if (bprop != null)
            {
              Resource tbp = typModel.createResource("TYP://" + bns + "#" +
                  localName);

              Statement ts = typModel.createStatement(tbp, RDFS.subPropertyOf,
                  tprop);

              typModel.add(ts);
            }
          }
        }
      }

      //copy all the subclass relations in current model to vpModel
      StmtIterator si = model.listStatements(new SelectorImpl(null,
          RDFS.subClassOf, (RDFNode)null));
      while (si.hasNext())
      {
        Statement stmt = si.nextStatement();
        Resource sub = stmt.getSubject();
        String suburi = sub.getURI();
        RDFNode objNode = stmt.getObject();
        Resource obj = (Resource) objNode;
        String objuri = obj.getURI();
        if (objuri == null)
        {
          continue;
        }
        String vpSuburi = "VPI://".concat(suburi);
        String vpObjuri = "VPI://".concat(objuri);
        Resource vSub = vpModel.getResource(vpSuburi);
        Resource vObj = vpModel.getResource(vpObjuri);
        if (vSub != null && vObj != null)
        {
          Statement s = vpModel.createStatement(vSub, RDFS.subClassOf, vObj);
          vpModel.add(s);
        }

      }

      //copy all the subproperty relations in current model to vpModel
      si = model.listStatements(new SelectorImpl(null, RDFS.subPropertyOf,
                                                 (RDFNode)null));
      while (si.hasNext())
      {
        Statement stmt = si.nextStatement();
        Resource sub = stmt.getSubject();
        String suburi = sub.getURI();
        RDFNode objNode = stmt.getObject();
        Resource obj = (Resource) objNode;
        String objuri = obj.getURI();
        if (objuri == null)
        {
          continue;
        }
        String vpSuburi = "VPI://".concat(suburi);
        String vpObjuri = "VPI://".concat(objuri);
        Resource vSub = vpModel.getResource(vpSuburi);
        Resource vObj = vpModel.getResource(vpObjuri);
        if (vSub != null && vObj != null)
        {
          Statement s = vpModel.createStatement(vSub, RDFS.subPropertyOf, vObj);
          vpModel.add(s);
        }
      }

      //copy all the "type" relations in current model to typModel
      si = model.listStatements();
      while (si.hasNext())
      {
        Statement stmt = si.nextStatement();
        Resource sub = stmt.getSubject();
        String suburi = sub.getURI();
        RDFNode objNode = stmt.getObject();
        Resource prop = stmt.getPredicate();

        if (objNode.equals(RDFS.Class) || objNode.equals(RDF.Property)
            || objNode.equals(OWL.Class) || objNode.equals(OWL.DatatypeProperty)
            || objNode.equals(OWL.ObjectProperty) || prop.equals(RDFS.domain)
            || prop.equals(RDFS.label) || prop.equals(RDFS.range)
            || prop.equals(RDFS.comment) || prop.equals(RDFS.subPropertyOf)
            || prop.equals(RDFS.subClassOf))
        {
          continue;
        }
        if (suburi == null)
        {
          continue;
        }

      //all the property hierarchies need to be copied onto the type model too

        if (prop.equals(RDF.type)) {
          Resource obj = (Resource) objNode;
          String objuri = obj.getURI();
          objuri = "TYP://".concat(objuri);

          Resource vObj = typModel.getResource(objuri);
          Resource tySub = typModel.getResource(suburi);
          if (vObj != null)
          {
            Statement s = typModel.createStatement(tySub, RDF.type, vObj);
            typModel.add(s);
          }
        }
        else
        {
          String propuri = prop.getURI();
          String tprop = "TYP://".concat(propuri);
          Property typrop = typModel.getProperty(tprop);
          Statement st = typModel.createStatement(sub, typrop, objNode);
          typModel.add(st);
        }
      }

      //Adding data in the VP Model to Global Model
      globalModel.add(vpModel);

      //Adding data in the Type Model to Global Model
      globalModel.add(typModel);

      System.out.println(uri);
      hasVP.add(uri);

      // printing out the RDF files for VPI, Type Graph and Global model

      vpModel.write(new FileWriter("vpi.rdf-" + pre));
      typModel.write(new FileWriter("typ.rdf-" + pre));
      globalModel.write(new FileWriter("global.rdf"));

    }

    catch (Exception e) {
      e.printStackTrace();
    }
  }

  /**
   * Submit a query to the VPI, using the default perspective.
   * @param q
   * @return
   */
  public ResultSet issueQuery(Query q) {

    List tp = q.getTriplePatterns();
    int size = tp.size();
    for (int i = 0; i < size; i++) {
      Triple t = (Triple) tp.get(i);
      Node nsub = alterNode(t.getSubject());
      Node npre = alterNode(t.getPredicate());
      Node nobj = alterNode(t.getObject());
      Triple nt = t.create(nsub, npre, nobj);
      tp.remove(t);
      tp.add(i, nt);
      System.out.println(nt);
    }
    return issueQueryWithoutVP(q);
  }

  /**
   * Submit a query to the VPI, using the default perspective.
   * @param q
   * @return
   */
  public ResultSet issueQueryWithoutVP(Query q) {
    q.setSource(globalModel);
    ResultSet rs = new ResultSet();
    QueryExecution qe = new QueryEngine(q);
    QueryResults results = qe.exec();
    int n = 0;
    System.out.println("Results:");
    while (results.hasNext()) {
      ResultBinding res = (ResultBinding) results.next();
      System.out.println(res.toString());
      String[] ss = {
          res.toString()};
      rs.addResult(new Result(ss));
      n++;
    }
    System.out.println(n + " answers");
    results.close();

    return rs;
  }

  /**
   * Submit a query to the VPI, specifing the perspective.
   * @param q
   * @param perspect
   * @return
   */
  public ResultSet issueQuery(Query q, String perspect) {
    //now we don't import anything, so we don't need perspectives.
    return null;
  }

  /**
   * Set the default perspective ontology of queries.
   * @param uri
   */
  public void setDefaultPerspective(String uri) {
    defaultPerspective = uri;
  }

  /**
   * Get the default perspective ontology of queries.
   * @return
   */
  public String getDefaultperspective() {
    return defaultPerspective;
  }

  /**
   * Alter the node to VP's uri if it has a VP.
   * @param n
   * @return
   */
  private Node alterNode(Node n) {
    if (n.isURI()) {
      String uri = n.getURI();
      String ns = uri.split("#")[0];
      if (hasVP.contains(ns)) {
        return n.createURI("VPI://".concat(uri));
      }
    }
    return n;
  }

  /**
   * Get the urls of the loadedDocuments
   * @return
   */
  public Object[] getLoadedDocuments() {
    Set set = loadedDocuments.keySet();
    return set.toArray();
  }

  private OntDocumentManager documentManager;
  private Hashtable loadedDocuments;
  private String defaultPerspective;
  private OntModel globalModel;
  private Vector hasVP;
}
